Unit TArray;

{Freeware Version 0.1

Dietmar Reinwald
Email: Dietmar.Reinwald@t-online.de
CompuServe: 101761,335

The 16-bit world is just alive. So I thought a lot about arrays > 64 KByte.
In my program BasToPas.exe I develop THugeMem from the class TMemoryStream.
Many thanks to the articel of Arne Schaeper "Gute Zeiten, schlechte Zeiten"
( "Good times, bad times") in c't 11/96 about the performance of Delphi-programs.
He presents the solution of arrays greater than 64 KByte.

I want to use this code (for all standard types) and two dimensional arrays in the
programm BasToPas.exe Version 0.6. (Freeware)

I take no responsibility if you use this code.
The use of this code is on your own risk and responsibility.

If you find errors please contact me.

Example:

procedure Test;
var
   myArray : TDoubleArray;
begin
   myArray := TDoubleArray.Create(15000);
       'This ceates an Double-array with index 0 to 14999
   myArray[10000]:=10000.0;
   myArray.ReDimPreserve(30000);
       'This creates an Double-array with index 0 to 29999 with preserving the old values
   if myArray[10000]<>10000.0 then
      ShowMessage('An error occurs');
   myArray.Free;

end;
}


interface
uses
    WinTypes, WinProcs, Classes;

type

      DoublePtr = ^Double;
      IntegerPtr = ^Integer;
      LongIntPtr = ^LongInt;
      SinglePtr = ^Single;
      ExtendedPtr = ^Extended;

TBaseArray = Class( TObject)
  protected
      myHandle :THandle;
      myPointer :Pointer;
      PosSize :LongInt;
      Itemsize :LongInt;
      procedure DisposeArray;
      procedure AllocArray(Wieviel :LongInt);
  public
      constructor Create(GetItemSize:integer; Wieviel :LongInt);
      destructor Destroy;override;
      function UBound:LongInt;
      procedure ReDimPreserve(NewSize :LongInt);
      procedure ReDim(NewSize :LongInt);
End;

TIntegerArray = Class( TBaseArray)
  public
      constructor Create(Wieviel :LongInt);
      Procedure SetData( index: LongInt; wert: Integer );
      Function GetData(index: LongInt): Integer;
      property Data[ Index:LongInt ]: Integer
         read GetData write SetData; default;
End;

TDoubleArray = Class( TBaseArray)
  public
      constructor Create(Wieviel :LongInt);
      Procedure SetData( index: LongInt; wert: Double );
      Function GetData(index: LongInt): Double;
      property Data[ Index:LongInt ]: double
         read GetData write SetData; default;
End;

TExtendedArray = Class( TBaseArray)
  public
      constructor Create(Wieviel :LongInt);
      Procedure SetData( index: LongInt; wert: Extended );
      Function GetData(index: LongInt): Extended;
      property Data[ Index:LongInt ]: Extended
         read GetData write SetData; default;
End;


implementation

      procedure _AHIncr; far; external 'Kernel' Index 114;

{------------ BaseVektor --------------------}

function TBaseArray .UBound:LongInt;
{We always think we have an array 0..PosSize-1}
begin
     UBound:=PosSize-1;
end;

procedure TBaseArray.DisposeArray;
begin
     if not (myHandle=0) then
     begin
       repeat
       until not GlobalUnlock(myHandle);
     end;
     myHandle:=GlobalFree(myHandle);
     myHandle:=0;
     myPointer:=nil;
end;


destructor TBaseArray.Destroy;
begin
     DisposeArray;
     inherited Destroy;
end;

procedure TBaseArray.AllocArray(Wieviel :LongInt);
begin
    myHandle:=GlobalAlloc(GMEM_MOVEABLE+GMEM_ZEROINIT,ItemSize*Wieviel);
    if myHandle<>0 then PosSize:=Wieviel;

end;

constructor TBaseArray.Create(GetItemSize:integer; Wieviel :LongInt);
{We want to have an array 0..Wieviel-1}
begin
    inherited Create;
    myHandle:=0;
    myPointer:=nil;
    ItemSize:=GetItemSize;
    AllocArray(Wieviel);
end;

procedure TBaseArray.ReDimPreserve(NewSize :LongInt);
{We want to have an array 0..NewSize-1}
begin
     myHandle:=GlobalReAlloc(myHandle, NewSize*ItemSize, GMEM_MOVEABLE + GMEM_ZEROINIT);
     if myHandle<>0 then PosSize:=NewSize;
end;

procedure TBaseArray.ReDim(NewSize :LongInt);
{We want to have an array 0..NewSize-1}
begin
     DisposeArray;
     AllocArray(NewSize);
end;

{--------------------------------- Integer ----------------------------------------------}
constructor TIntegerArray.Create(Wieviel :LongInt);
{We want to have an array 0..NewSize-1}
begin
      inherited Create(sizeOf(Integer), Wieviel);
end;


function TIntegerArray.GetData(index:longint):Integer ;
var
      mySize, Offset, Selector :LongInt;
      good :boolean;
      myIntegerPtr :IntegerPtr;
begin
      Result:=0;
      If index>Ubound then
          ReDimPreserve(index+1);

      myPointer:=GlobalLock(myHandle);
      mySize:=ItemSize*index;

      Selector:=mySize div 65636;
      Offset:=mySize mod 65636;
      if Selector=0 then
         myIntegerPtr:=Ptr(Seg(myPointer^), Offset)
      else
         myIntegerPtr:=Ptr(Seg(myPointer^)+Selector*(Ofs(_AHIncr)), Offset);

      Result:=myIntegerPtr^;
      good:=GlobalUnLock(myHandle);
end;

procedure TIntegerArray.SetData(index:longint; wert:Integer);
var
      mySize, Offset, Selector :LongInt;
      good :boolean;
      myIntegerPtr :IntegerPtr;
begin
      If index>Ubound then
          ReDimPreserve(index+1);

      myPointer:=GlobalLock(myHandle);
      mySize:=ItemSize*index;

      Selector:=mySize div 65636;
      Offset:=mySize mod 65636;
      if Selector=0 then
         myIntegerPtr:=Ptr(Seg(myPointer^), Offset)
      else
         myIntegerPtr:=Ptr(Seg(myPointer^)+Selector*(Ofs(_AHIncr)), Offset);

      myIntegerPtr^:=wert;
      good:=GlobalUnLock(myHandle);
end;

{--------------------------------- Double ----------------------------------------------}
constructor TDoubleArray.Create(Wieviel :LongInt);
{We want to have an array 0..NewSize-1}
begin
      inherited Create(sizeOf(double), Wieviel);
end;


function TDoubleArray.GetData(index:longint):double ;
var
      mySize, Offset, Selector :LongInt;
      good :boolean;
      myDoublePtr :DoublePtr;
begin
      Result:=0.0;
      If index>Ubound then
          ReDimPreserve(index+1);

      myPointer:=GlobalLock(myHandle);
      if myPointer=nil then exit;
      mySize:=ItemSize*index;

      Selector:=mySize div 65636;
      Offset:=mySize mod 65636;
      if Selector=0 then
         myDoublePtr:=Ptr(Seg(myPointer^), Offset)
      else
         myDoublePtr:=Ptr(Seg(myPointer^)+Selector*(Ofs(_AHIncr)), Offset);

      Result:=myDoublePtr^;
      good:=GlobalUnLock(myHandle);
end;

procedure TDoubleArray.SetData(index:longint; wert:double);
var
      mySize, Offset, Selector :LongInt;
      good :boolean;
      myDoublePtr :DoublePtr;
begin
      If index>Ubound then
          ReDimPreserve(index+1);

      myPointer:=GlobalLock(myHandle);
      if myPointer=nil then exit;
      mySize:=ItemSize*index;

      Selector:=mySize div 65636;
      Offset:=mySize mod 65636;
      if Selector=0 then
         myDoublePtr:=Ptr(Seg(myPointer^), Offset)
      else
         myDoublePtr:=Ptr(Seg(myPointer^)+Selector*(Ofs(_AHIncr)), Offset);

      myDoublePtr^:=wert;
      good:=GlobalUnLock(myHandle);
end;

{--------------------------------- Extended ----------------------------------------------}
constructor TExtendedArray.Create(Wieviel :LongInt);
{We want to have an array 0..NewSize-1}
begin
      inherited Create(sizeOf(Extended), Wieviel);
end;


function TExtendedArray.GetData(index:longint):Extended ;
var
      mySize, Offset, Selector :LongInt;
      good :boolean;
      myExtendedPtr :ExtendedPtr;
begin
      Result:=0.0;
      If index>Ubound then
          ReDimPreserve(index+1);

      myPointer:=GlobalLock(myHandle);
      mySize:=ItemSize*index;

      Selector:=mySize div 65636;
      Offset:=mySize mod 65636;
      if Selector=0 then
         myExtendedPtr:=Ptr(Seg(myPointer^), Offset)
      else
         myExtendedPtr:=Ptr(Seg(myPointer^)+Selector*(Ofs(_AHIncr)), Offset);

      Result:=myExtendedPtr^;
      good:=GlobalUnLock(myHandle);
end;

procedure TExtendedArray.SetData(index:longint; wert:Extended);
var
      mySize, Offset, Selector :LongInt;
      good :boolean;
      myExtendedPtr :ExtendedPtr;
begin
      If index>Ubound then
          ReDimPreserve(index+1);

      myPointer:=GlobalLock(myHandle);
      mySize:=ItemSize*index;

      Selector:=mySize div 65636;
      Offset:=mySize mod 65636;
      if Selector=0 then
         myExtendedPtr:=Ptr(Seg(myPointer^), Offset)
      else
         myExtendedPtr:=Ptr(Seg(myPointer^)+Selector*(Ofs(_AHIncr)), Offset);

      myExtendedPtr^:=wert;
      good:=GlobalUnLock(myHandle);
end;

end.